# Copyright (C) 2014 Julian Metzler
# See the LICENSE file for the full license.

"""
Message types for AM03127-based LED signs
"""

class RawMessage(object):
	"""
	A raw datagram that can be sent to the sign, with all the various properties
	This class is usually instantiated by a LEDSign instance which fills in the ID field
	"""
	
	BASE_FORMAT = "<ID%(id)02X>%(data)s%(checksum)02X<E>"
	
	def __init__(self, id, data):
		self.format_data = {
			'id': id,
			'data': data,
			'checksum': 0
		}
	
	def calculate_checksum(self):
		for char in self.format_data['data']:
			self.format_data['checksum'] ^= ord(char)
	
	def render(self):
		self.calculate_checksum()
		return self.BASE_FORMAT % self.format_data

class SetIDMessage(RawMessage):
	"""
	A datagram to set the sign's ID. Differs from all other datagram types in that
	it doesn't have a checksum or ID field
	"""
	
	BASE_FORMAT = "<ID><%(id)02X><E>"
	
	def __init__(self, id):
		self.format_data = {
			'id': id
		}
	
	def set_id(self, id):
		"""
		For compatibility only
		"""
		
		return
	
	def render(self):
		return self.BASE_FORMAT % self.format_data

class BaseMessage(object):
	"""
	A template for datagrams that already includes the raw formatting
	As with RawMessage, the ID field is filled in by the LEDSign instance
	the moment it renders the message
	"""
	
	TEMPLATE = ""
	
	def __init__(self, id = 0, **data):
		self.id = id
		self.format_data = data
	
	def set_id(self, id):
		self.id = id
	
	def render(self):
		self.formatted_data = self.TEMPLATE % self.format_data
		msg = RawMessage(id = self.id, data = self.formatted_data)
		return msg.render()

class SetClockMessage(BaseMessage):
	"""
	Set the clock in the LED sign
	"""
	
	TEMPLATE = "<SC>%(year)02i%(weekday)02i%(month)02i%(day)02i%(hour)02i%(minute)02i%(second)02i"

class SendPageMessage(BaseMessage):
	"""
	Send a message to a page
	"""
	
	TEMPLATE = "<L%(line)i><P%(page)c><F%(lead)c><M%(method)c><W%(wait)c><F%(lag)c>%(content)s"

class SendScheduleMessage(BaseMessage):
	"""
	Send a schedule to the LED sign
	"""
	
	TEMPLATE = "<T%(schedule)c>%(startyear)02i%(startmonth)02i%(startday)02i%(starthour)02i%(startminute)02i%(endyear)02i%(endmonth)02i%(endday)02i%(endhour)02i%(endminute)02i%(pages)s"

class SendGraphicMessage(BaseMessage):
	"""
	Send a graphic to the LED sign
	"""
	
	TEMPLATE = "<G%(page)c%(block)i>%(data)s"

class DeletePageMessage(BaseMessage):
	"""
	Delete a page
	"""
	
	TEMPLATE = "<DL%(line)iP%(page)c>"

class DeleteScheduleMessage(BaseMessage):
	"""
	Delete a schedule
	"""
	
	TEMPLATE = "<DT%(schedule)c>"

class DeleteAllMessage(BaseMessage):
	"""
	Delete all data
	"""
	
	TEMPLATE = "<D*>"

class SetRunPageMessage(BaseMessage):
	"""
	Set the default page to run if no schedule is active
	"""
	
	TEMPLATE = "<RP%(page)c>"

class SetBrightnessMessage(BaseMessage):
	"""
	Set the brightness of the LED sign
	"""
	
	TEMPLATE = "<B%(level)c>"

class SendCharacterMessage(BaseMessage):
	"""
	Send a special character definition
	"""
	
	TEMPLATE = "<F%(font)c%(code)02X>%(data)s"

class ResetCharacterTableMessage(BaseMessage):
	"""
	Revert to the factory default special character table
	"""
	
	TEMPLATE = "<DU>"

class PageContent(object):
	"""
	A subclass representing the content of a page used in SendPageMessage
	"""
	
	def __init__(self, data):
		self.data = data
	
	def render(self):
		rendered_message = ""
		for part in self.data:
			if type(part) is not dict:
				rendered_message += part
				continue
			
			for key, value in part.iteritems():
				if key == 'text':
					rendered_message += value
					continue
				
				try:
					func = getattr(self, '_get_%s_tag' % key)
				except AttributeError:
					continue
				if type(value) is dict:
					tag = func(**value)
				else:
					tag = func(value)
				rendered_message += tag
		return rendered_message
	
	@classmethod
	def _get_font_tag(cls, font):
		if font == 'normal':
			char = "A"
		elif font == 'bold':
			char = "B"
		elif font == 'narrow':
			char = "C"
		elif font == 'large':
			char = "D"
		elif font == 'long':
			char = "E"
		else:
			char = "A"
		return "<A%c>" % char
	
	@classmethod
	def _get_bell_tag(cls, bell):
		try:
			bell = float(bell)
		except (TypeError, ValueError):
			bell = 1.0
		
		char = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"[int(bell * 2 + 0.5) - 1]
		return "<B%c>" % char
	
	@classmethod
	def _get_color_tag(cls, color):
		if color == 'dim-red':
			char = "A"
		elif color == 'red':
			char = "B"
		elif color == 'bright-red':
			char = "C"
		elif color == 'dim-green':
			char = "D"
		elif color == 'green':
			char = "E"
		elif color == 'bright-green':
			char = "F"
		elif color == 'dim-orange':
			char = "G"
		elif color == 'orange':
			char = "H"
		elif color == 'bright-orange':
			char = "I"
		elif color == 'yellow':
			char = "J"
		elif color == 'lime':
			char = "K"
		elif color == 'inverted-red':
			char = "L"
		elif color == 'inverted-green':
			char = "M"
		elif color == 'inverted-orange':
			char = "N"
		elif color == 'red-on-green':
			char = "P"
		elif color == 'green-on-red':
			char = "Q"
		elif color == 'ryg':
			char = "R"
		elif color == 'rainbow':
			char = "S"
		else:
			char = "B"
		return "<C%c>" % char
	
	@classmethod
	def _get_graphic_tag(cls, page, block):
		# Optimize this for ease of use!
		return "<G%c%i>" % (page, block)
	
	@classmethod
	def _get_character_tag(cls, code):
		try:
			code = int(code)
		except (TypeError, ValueError):
			code = 0
		return "<U%02X>" % code
	
	@classmethod
	def _get_column_tag(cls, column):
		try:
			column = int(column)
		except (TypeError, ValueError):
			column = 0
		return "<N%02X>" % column
	
	@classmethod
	def _get_datetime_tag(cls, type):
		if type == 'date':
			char = "D"
		elif type == 'time':
			char = "T"
		else:
			char = "T"
		return "<K%c>" % char


# Copyright (C) 2014 Julian Metzler
# See the LICENSE file for the full license.

"""
Communication with AM03127-based LED signs
"""

from .messages import *
import datetime
import serial
import time

class SerialCommunicator(object):
	"""
	Manager class for serial communication with the LED Sign
	"""
	
	BYTESIZE = serial.EIGHTBITS
	PARITY = serial.PARITY_NONE
	STOPBITS = serial.STOPBITS_ONE
	
	PROCESSING_TIME = 0.5 # How long we should wait between sending a command and reading the response
	
	def __init__(self, port, baudrate, timeout):
		self.port = port
		self.baudrate = baudrate
		self.timeout = timeout
		self.init_comm()
	
	def init_comm(self):
		self.device = serial.serial_for_url(self.port,
			baudrate = self.baudrate,
			bytesize = self.BYTESIZE,
			parity = self.PARITY,
			stopbits = self.STOPBITS,
			timeout = self.timeout,
			writeTimeout = self.timeout
		)
	
	def blocking_write(self, data):
		"""
		Perform a write operation and wait until it's likely to have finished
		"""
		
		num_bytes = self.device.write(data)
		return num_bytes
	
	def send_command(self, data):
		"""
		Send data to the device, wait for it to process the data and read the response
		"""
		
		self.blocking_write(data)
		time.sleep(self.PROCESSING_TIME)
		response = self.device.read(self.device.inWaiting())
		return response

class LEDSign(object):
	"""
	LED Sign class, for actually interfacing with the sign
	"""
	
	EFFECT_IMMEDIATE = "A"
	EFFECT_XOPEN = "B"
	EFFECT_CURTAIN_UP = "C"
	EFFECT_CURTAIN_DOWN = "D"
	EFFECT_SCROLL_LEFT = "E"
	EFFECT_SCROLL_RIGHT = "F"
	EFFECT_VOPEN = "G"
	EFFECT_VCLOSE = "H"
	EFFECT_SCROLL_UP = "I"
	EFFECT_SCROLL_DOWN = "J"
	EFFECT_HOLD = "K"
	EFFECT_SNOW = "L"
	EFFECT_TWINKLE = "M"
	EFFECT_BLOCK_MOVE = "N"
	EFFECT_RANDOM = "P"
	EFFECT_HELLO_WORLD = "Q"
	EFFECT_WELCOME = "R"
	EFFECT_AMPLUS = "S"
	
	METHOD_NORMAL = 0x01
	METHOD_BLINKING = 0x02
	METHOD_SONG_1 = 0x03
	METHOD_SONG_2 = 0x04
	METHOD_SONG_3 = 0x05
	
	SPEED_FAST = 0x40
	SPEED_MEDIUM = 0x50
	SPEED_SLOW = 0x60
	SPEED_SLOWEST = 0x70
	
	def __init__(self, port = None, baudrate = 9600, timeout = None, id = 1):
		self.id = id
		self.comm = SerialCommunicator(
			port = port,
			baudrate = baudrate,
			timeout = timeout
		)
	
	def _get_page_char(self, page):
		return "ABCDEFGHIJKLMNOPQRSTUVWXYZ"[page - 1]
	
	def _get_schedule_char(self, schedule):
		return "ABCDE"[schedule - 1]
	
	def _get_wait_char(self, duration):
		if duration < 1.0:
			char = "A"
		else:
			char = "BCDEFGHIJKLMNOPQRSTUVWXYZ"[int(duration) - 1]
		return char
	
	def send_raw(self, data, expected_response = "ACK"):
		"""
		Send the given data to the sign, read the response, decide whether the sign
		acknowledges the data and return a boolean indicating success or failure
		"""
		
		response = self.comm.send_command(data)
		# print repr(response)
		
		if response == expected_response:
			success = True
		else:
			success = False
		
		return success
	
	def send_message(self, message):
		"""
		Send a message instance to the sign
		"""
		
		message.set_id(self.id)
		# print message.render()
		
		if isinstance(message, SetIDMessage):
			expected_response = "%02X" % message.format_data['id']
		else:
			expected_response = "ACK"
		
		success = self.send_raw(message.render(), expected_response)
		return success
	
	def set_id(self, id):
		"""
		Set the sign's ID
		"""
		
		msg = SetIDMessage(id = id)
		return self.send_message(msg)
	
	def set_clock(self, timedata = None):
		"""
		Set the sign's internal clock
		"""
		
		if timedata is None:
			timedata = datetime.datetime.now()
		
		msg = SetClockMessage(
			year = timedata.strftime("%y"),
			weekday = timedata.isoweekday(),
			month = timedata.month,
			day = timedata.day,
			hour = timedata.hour,
			minute = timedata.minute,
			second = timedata.second
		)
		return self.send_message(msg)
	
	def send_page(self, content, page = "A", line = 1, lead = EFFECT_SCROLL_LEFT, speed = SPEED_MEDIUM, method = METHOD_NORMAL, wait = 2.0, lag = EFFECT_SCROLL_LEFT):
		"""
		Send text to a page
		"""
		
		if not isinstance(content, PageContent):
			content = PageContent(content)
		
		if type(page) not in (str, unicode):
			page = self._get_page_char(page)
		
		if type(wait) not in (str, unicode):
			wait = self._get_wait_char(wait)
		
		if type(method) not in (str, unicode):
			method = chr(speed + method)
		
		msg = SendPageMessage(
			line = line,
			page = page.upper(),
			lead = lead,
			method = method,
			wait = wait.upper(),
			lag = lag,
			content = content.render()
		)
		return self.send_message(msg)
	
	def send_schedule(self, schedule = "A", start = None, end = None, pages = "A", recurring = False):
		"""
		Send a schedule
		"""
		
		if start is None:
			start = datetime.datetime(
				year = 2000,
				month = 1,
				day = 1
			)
		
		if end is None:
			end = datetime.datetime(
				year = 2099,
				month = 12,
				day = 31
			)
		
		if recurring:
			startyear = 0
			startmonth = 0
			startday = 0
			starthour = start.hour
			startminute = start.minute
			endyear = 0
			endmonth = 0
			endday = 0
			endhour = end.hour
			endminute = end.minute
		else:
			startyear = int(start.strftime("%y"))
			startmonth = start.month
			startday = start.day
			starthour = start.hour
			startminute = start.minute
			endyear = int(end.strftime("%y"))
			endmonth = end.month
			endday = end.day
			endhour = end.hour
			endminute = end.minute
		
		if type(schedule) not in (str, unicode):
			schedule = self._get_schedule_char(schedule)
		
		msg = SendScheduleMessage(
			schedule = schedule.upper(),
			startyear = startyear,
			startmonth = startmonth,
			startday = startday,
			starthour = starthour,
			startminute = startminute,
			endyear = endyear,
			endmonth = endmonth,
			endday = endday,
			endhour = endhour,
			endminute = endminute,
			pages = pages.upper()
		)
		return self.send_message(msg)
	
	def delete_page(self, page, line):
		"""
		Delete a page
		"""
		
		if type(page) not in (str, unicode):
			page = self._get_page_char(page)
		
		msg = DeletePageMessage(
			page = page,
			line = line
		)
		return self.send_message(msg)
	
	def delete_schedule(self, schedule):
		"""
		Delete a schedule
		"""
		
		if type(schedule) not in (str, unicode):
			schedule = self._get_schedule_char(schedule)
		
		msg = DeleteScheduleMessage(
			schedule = schedule
		)
		return self.send_message(msg)
	
	def delete_all(self):
		"""
		Delete all data
		"""
		
		msg = DeleteAllMessage()
		return self.send_message(msg)
	
	def set_run_page(self, page):
		"""
		Set the default run page
		"""
		
		if type(page) not in (str, unicode):
			page = self._get_page_char(page)
		
		msg = SetRunPageMessage(
			page = page
		)
		return self.send_message(msg)
	
	def set_brightness(self, level):
		"""
		Set the LED brightness
		"""
		
		if type(level) not in (str, unicode):
			level = "ABCDD"[4 - int(divmod(level, 0.25)[0])]
		
		msg = SetBrightnessMessage(
			level = level
		)
		return self.send_message(msg)
	
	def reset_character_table(self):
		"""
		Reset the character table to the factory default
		"""
		
		msg = ResetCharacterTableMessage()
		return self.send_message(msg)


# Copyright (C) 2014 Julian Metzler
# See the LICENSE file for the full license.

name = "pyLEDSign"
version = "2.0.1"
description = "A LED message board library for Python"
license = "AGPLv3"
author = "Julian Metzler"
author_email = "contact@mezgrman.de"
requires = ['bbcode']
url = "https://github.com/Mezgrman/pyLEDSign"
keywords = "led sign message board effect library wrapper serial scrolling text"



#!/usr/bin/env python
# Copyright (C) 2014 Julian Metzler
# See the LICENSE file for the full license.

"""
EXAMPLE SCRIPT: Interactive shell for AM03127-based LED signs
"""

import argparse
import cmd
import ledsign

class InteractiveShell(cmd.Cmd):
	"""
	The "interpreter" for the LED sign shell
	"""
	
	prompt = ">>> "
	
	def do_pages(self, pages):
		success = self.sign.send_schedule(
			schedule = "A",
			pages = pages
		)
		print "Pages set!" if success else "Failed to set pages"
	
	def do_page(self, command):
		cmdparts = command.split()
		page = cmdparts[0]
		text = " ".join(cmdparts[1:])
		
		content = ledsign.am03127.PageContentBBCodeParser().render(text)
		
		success = self.sign.send_page(
			page = page,
			lead = self.settings['lead'],
			speed = self.settings['speed'],
			method = self.settings['method'],
			wait = self.settings['wait'],
			lag = self.settings['lag'],
			content = content
		)
		print "Page sent!" if success else "Failed to send page"

def main():
	parser = argparse.ArgumentParser(description = "Interactive AM03127 LED sign shell")
	parser.add_argument('-d', '--device',
		default = "/dev/ttyUSB0",
		help = "Serial device to use")
	
	parser.add_argument('-b', '--baudrate',
		type = int,
		choices = (1200, 2400, 4800, 9600, 19200),
		default = 9600,
		help = "Baudrate for the serial port")
	
	parser.add_argument('-i', '--id',
		type = int,
		default = 1,
		help = "ID of the LED sign")
	
	parser.add_argument('-s', '--speed',
		choices = ('slowest', 'slow', 'medium', 'fast'),
		default = 'medium',
		help = "Effect speed")
	
	parser.add_argument('-m', '--method',
		choices = ('normal', 'blinking', 'song_1', 'song_2', 'song_3'),
		default = 'normal',
		help = "Display effect while waiting")
	
	parser.add_argument('-w', '--wait',
		type = float,
		default = 2.0,
		help = "Time to show the text until it disappears")
	
	parser.add_argument('-lead', '--lead',
		choices = ('immediate', 'xopen', 'curtain_up', 'curtain_down', 'scroll_left', 'scroll_right', 'vopen', 'vclose', 'scroll_up', 'scroll_down', 'hold', 'snow', 'twinkle', 'block_move', 'random', 'hello_world', 'welcome', 'amplus'),
		default = 'scroll_left',
		help = "Leading effect")
	
	parser.add_argument('-lag', '--lag',
		choices = ('immediate', 'xopen', 'curtain_up', 'curtain_down', 'scroll_left', 'scroll_right', 'vopen', 'vclose', 'scroll_up', 'scroll_down', 'hold'),
		default = 'scroll_left',
		help = "Lagging effect")
	
	args = parser.parse_args()
	
	sign = ledsign.am03127.LEDSign(
		port = args.device,
		baudrate = args.baudrate,
		timeout = None,
		id = args.id
	)
	
	settings = {
		'speed': getattr(sign, "SPEED_%s" % args.speed.upper()),
		'method': getattr(sign, "METHOD_%s" % args.method.upper()),
		'wait': args.wait,
		'lead': getattr(sign, "EFFECT_%s" % args.lead.upper()),
		'lag': getattr(sign, "EFFECT_%s" % args.lag.upper())
	}
	
	shell = InteractiveShell()
	shell.sign = sign
	shell.settings = settings
	try:
		shell.cmdloop()
	except KeyboardInterrupt:
		print

if __name__ == "__main__":
	main()


# Copyright (C) 2014 Julian Metzler
# See the LICENSE file for the full license.

"""
Message parsers for easier human interfacing
"""

from .messages import *
import bbcode

class BaseParser(object):
	"""
	A general parser that renders a given class
	"""
	
	TARGET = object
	
	def __init__(self):
		pass
	
	def parse(self, data):
		"""
		The actual parsing takes place here, this must be subclassed and the kwargs returned
		"""
		
		return {}
	
	def render(self, data):
		target_kwargs = self.parse(data)
		instance = self.TARGET(**target_kwargs)
		return instance

class PageContentBBCodeParser(BaseParser):
	"""
	Parse page content from BBCode
	"""
	
	TARGET = PageContent
	
	def __init__(self):
		def _bb_dummy(tagname, value, options, parent, context):
			return
		
		self.bbcode_parser = bbcode.Parser()
		self.bbcode_parser.add_formatter('font', _bb_dummy, standalone = True)
		self.bbcode_parser.add_formatter('bell', _bb_dummy, standalone = True)
		self.bbcode_parser.add_formatter('color', _bb_dummy, standalone = True)
		self.bbcode_parser.add_formatter('graphic', _bb_dummy, standalone = True)
		self.bbcode_parser.add_formatter('char', _bb_dummy, standalone = True)
		self.bbcode_parser.add_formatter('column', _bb_dummy, standalone = True)
		self.bbcode_parser.add_formatter('date', _bb_dummy, standalone = True)
		self.bbcode_parser.add_formatter('time', _bb_dummy, standalone = True)
	
	def parse(self, data):
		target_data = []
		tokens = self.bbcode_parser.tokenize(data)
		for token_type, tag_name, tag_options, token_text in tokens:
			if token_type == self.bbcode_parser.TOKEN_TAG_START:
				if tag_name in ('date', 'time'):
					key = 'datetime'
					value = tag_name
				else:
					key = tag_name
					value = tag_options.get(tag_name)
				
				target_data.append({
					key: value
				})
			elif token_type == self.bbcode_parser.TOKEN_DATA:
				target_data.append(token_text)
		target_kwargs = {'data': target_data}
		return target_kwargs